/* sonnenpos.cc			letzte Aenderung: 23.6.2015 */
#define VERSION "Version 0.0"
/*
Uebersetzen auf Unix (Linux):
> make  ;siehe makefile

 Kurzbeschreibung: relativ genaue Berechnung der Sonnenposition

History:
22.6.2015	Erstellung (RP)
*/

#define MITREFRAKTION  //auskommentieren fuer Berechnungen ohne Refraktion

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>
#ifndef PI
const double PI=(atan(1.0)*4.0);
#endif
#ifndef ZWEIPI
#define ZWEIPI (2.0*PI)
#endif
#ifndef GRAD
#define GRAD (PI/180.0)
#endif
#ifndef WINKELMINUTE
#define WINKELMINUTE (GRAD/60.0)
#endif

/************************* Vordeklarationen ***************************/

/**************** Routinen zur Parameterauswertung ********************/
#define MAXARG 4
static char argflag[128];
void setargflags(const char *s)
{
 int c;
 while(c= *s++)
  {if(c>='a' && c<='z')  c -= 'a'-'A';
   argflag[c&127]=1;
  }
}

/************** Formel zum Sonnenposition berechnen *******************/
// JD-Formel: http://www.chem.uzh.ch/pfister/formelsammlung.html#mjd
// weitere Formeln: https://de.wikipedia.org/wiki/Sonnenstand

double MJD(int jahr,int monat,int tag,int stunde,int minute,int sekunde) //Zeit in UT
{
 int y,m,B,d=tag;
 double h=stunde+minute/60.0+sekunde/3600.0;
 if(monat<=2) {y=jahr-1; m=monat+13;}
 else         {y=jahr;   m=monat+1;}
 if(y<1582 || (y==1582 && (monat<10 || (monat==10 && tag<=4))))
            B=(y+4716)/4-1181; //bis 4.Okt.1582
 else       B=y/400-y/100+y/4; //ab 15.Okt.1582
 return 365.0*y-679004+B+int(30.6*m)+d+h/24.0;
}

void sonnenposition_berechnen(int jahr,int monat,int tag, //Datum und
			      int stunde,int minute,int sekunde, //Zeit in UT
			      double lat,double lon, //Geographische Koordinaten in Grad
			      double *hoehe,double *azimut, //Rueckgabewerte in Rad
			      int refraktflag=1)
{
 //printf("%d-%d-%d %d:%02d:%02d\n",jahr,monat,tag,stunde,minute,sekunde);//test
 //double n = JD(jahr,monat,tag,stunde,minute,sekunde) - 2451545.0;
 double n = MJD(jahr,monat,tag,stunde,minute,sekunde) - 51544.5; //in MJD ist 2400000.5 schon subrahiert
 //printf("JD: n=%f\n",n);//test
 double L = 280.460*GRAD + 0.9856474*GRAD*n; //mittlere Laenge
 double g = 357.528*GRAD + 0.9856003*GRAD*n; //mittlere Anomalie
 //eventuell noch L und g in den Bereich von 0 bis 2Pi bringen:
 while(L>=ZWEIPI) L-=ZWEIPI;
 while(g>=ZWEIPI) g-=ZWEIPI;
 //printf("L=%f Grad, g=%f Grad\n",L/GRAD,g/GRAD);//test
 double Lambda = L + 1.915*GRAD*sin(g) + 0.01997*GRAD*sin(2*g); //ekliptikale Laenge
 double epsilon = 23.439*GRAD - 0.0000004*GRAD*n; //Schiefe der Ekliptik
 double alfa = atan2(cos(epsilon)*sin(Lambda),cos(Lambda)); //Rektaszension
 double delta = asin(sin(epsilon)*sin(Lambda)); //Deklination
 //double T0 = (JD(jahr,monat,tag,0,0,0) - 2451545.0)/36525.0;
 double T0 = (MJD(jahr,monat,tag,0,0,0) - 51544.5)/36525.0;
 double T = stunde+minute/60.0+sekunde/3600.0;
 double thetahg = 6.697376 + 2400.05134*T0 + 1.002738*T;
 while(thetahg>=24.0) thetahg -= 24.0;
 double thetag = thetahg*15*GRAD; //mittlere Sternzeit
 double theta = thetag+lon*GRAD; //Stundenwinkel des Fruehlingspunkts
 double tau = theta-alfa; //Stundenwinkel der Sonne
 double phi = lat*GRAD; //Geographische Breite
 *azimut = atan2(sin(tau), cos(tau)*sin(phi) - tan(delta)*cos(phi));
 *hoehe = asin(cos(delta)*cos(tau)*cos(phi) + sin(delta)*sin(phi));
 //Refraktion beruecksichtigen:
#ifdef MITREFRAKTION
 double h= *hoehe/GRAD;
 if(refraktflag!=0 && h<45 && h > -45)
  {
   if(h<0) {if(h < -2)  h = -(h+2); else h=0;}
   double R=1.02/tan((h+10.3/(h+5.11))*GRAD); //Refraktion in Winkelminuten
   if(argflag['V']) printf("R=%f Winkelminuten\n",R);//test
   *hoehe += R*WINKELMINUTE;
  }
#endif
}

static int zeitzone=1;
static int monatstage[13]={0,31,28,31,30,31,30,31,31,30,31,30,31};

void umrechnung_zonenzeit_nach_ut(int& jahr,int& monat,int& tag,int& stunde)
{
 int schaltjahr=0;
 if((jahr%4)==0 && (jahr%100!=0 || jahr%400==0)) schaltjahr=1;
 stunde -= zeitzone;
 if(stunde>=24)
  {
   int max;
   stunde -= 24;
   if(schaltjahr==1 && monat==2) max=29;
   else max=monatstage[monat];
   if(++tag>max)
    {
     tag=1;
     if(++monat==13) {monat=1; ++jahr;}
    }
  }
 else if(stunde<0)
  {
   stunde += 24;
   if(--tag==0)
    {
     if(--monat==0) {tag=31; monat=12; --jahr;}
     else
      {
       if(schaltjahr==1 && monat==2) tag=29;
       else tag=monatstage[monat];
      }
    }
  }
}

/************************* Hauptprogramm ******************************/
int main(int argc,char *argv[])
{
 char datumeingabe[80],zeiteingabe[80];
 int i,j=0,c;
 int jahr,monat,tag,stunde,min,sek;
 double hoehe,azimut;
 //double lat=48.1, lon=11.6; //Voreinstellung: Muenchen
 double lat=47.367, lon=8.533; //Voreinstellung: Zuerich
 datumeingabe[0]=zeiteingabe[0]=0;
 if(argc<=0)
  {/* es wurde von WorkBench (GUI, Desktop, ...) gestartet */
   return 1;
  }
 else
  {/* es wurde von Shell (Terminal, Konsole, ...) gestartet */
   for(j=0,i=1;i<argc;i++)
    {if((c= *argv[i])=='?' || (c=='-' && !isdigit(argv[i][1]))) setargflags(argv[i]);
	 else	{if(++j==1)    sscanf(argv[i],"%lf",&lat);
	         else if(j==2) sscanf(argv[i],"%lf",&lon);
	         else if(j==3) strcpy(datumeingabe,argv[i]);
	         else if(j==4) strcpy(zeiteingabe,argv[i]);
	}	}
  }
 if(datumeingabe[0]==0) printf("fehlende Parameter\n");
 if(argflag['?'] || argflag['h'] || j<1 || j>MAXARG)
	{printf("sonnenpos  %s\n",VERSION);
	 printf("Anwendung: sonnenpos GeoBreite GeoLaenge Datum [Zeit]\n");
	 printf("  Geographische Breite und Laenge in Grad angeben\n");
	 printf("  Form von Datum und Zeit: 2015-6-22 15:34:00\n");
	 printf("  Die Zeit muss als Weltzeit (UT) angegeben werden\n");
	 printf("Beispiele:\n");
	 printf(" sonnenpos 48.1 11.6 2006-8-6 6:00:00\n");
	 printf(" sonnenpos 47.367 8.533 2015-6-23\n");
	 exit(0);
	}
 sscanf(datumeingabe,"%d%*c%d%*c%d",&jahr,&monat,&tag);
 if(zeiteingabe[0]!=0)
  {
   sscanf(zeiteingabe,"%d%*c%d%*c%d",&stunde,&min,&sek);
   sonnenposition_berechnen(jahr,monat,tag,stunde,min,sek,lat,lon,&hoehe,&azimut);
   printf("Hoehe=%f Grad  Azimut=%f Grad\n",hoehe/GRAD,azimut/GRAD);
   sonnenposition_berechnen(jahr,monat,tag,stunde,min,sek,lat,lon,&hoehe,&azimut,0);
   printf("ohne Refraktion: Hoehe=%f Grad  Azimut=%f Grad\n",hoehe/GRAD,azimut/GRAD);
  }
 else
  {
   int tagflag=0,maxflag=0;
   double hoehe_vorher,azimut_vorher=0;
   for(stunde=min=sek=0,hoehe_vorher= -1; stunde<24;)
    {
     sonnenposition_berechnen(jahr,monat,tag,stunde,min,sek,lat,lon,&hoehe,&azimut);
     if(hoehe>=0 && hoehe_vorher<0)
      {
       printf("Sonnenaufgang: %d:%02d:%02d  Azimut: %.2f\n",stunde,min,sek,azimut/GRAD);
       tagflag=1;
      }
     else if(hoehe<0 && hoehe_vorher>=0)
      {
       printf("Sonnenuntergang: %d:%02d:%02d  Azimut: %.2f\n",stunde,min,sek,azimut/GRAD);
       tagflag=0;
      }
     if(tagflag==1 && maxflag==0 && hoehe_vorher>hoehe)
      {
       printf("Hoechststand: %.2f Grad,  Azimut: %.2f Grad\n",hoehe_vorher/GRAD,azimut_vorher/GRAD);
       maxflag=1;
      }
     hoehe_vorher=hoehe;
     azimut_vorher=azimut;
     if(++sek==60) {sek=0; if(++min==60) {min=0; stunde++;}}
    }
   for(stunde=min=sek=0,hoehe_vorher= -1; stunde<24;)
    {
     sonnenposition_berechnen(jahr,monat,tag,stunde,min,sek,lat,lon,&hoehe,&azimut,0);
     if(hoehe>=0 && hoehe_vorher<0)
      {
       printf("Sonnenaufgang ohne Refraktion: %d:%02d:%02d  Azimut: %.2f\n",stunde,min,sek,azimut/GRAD);
       tagflag=1;
      }
     else if(hoehe<0 && hoehe_vorher>=0)
      {
       printf("Sonnenuntergang ohne Refraktion: %d:%02d:%02d  Azimut: %.2f\n",stunde,min,sek,azimut/GRAD);
       tagflag=0;
      }
     if(tagflag==1 && maxflag==0 && hoehe_vorher>hoehe)
      {
       printf("Hoechststand: %.2f Grad,  Azimut: %.2f Grad\n",hoehe_vorher/GRAD,azimut_vorher/GRAD);
       maxflag=1;
      }
     hoehe_vorher=hoehe;
     azimut_vorher=azimut;
     if(++sek==60) {sek=0; if(++min==60) {min=0; stunde++;}}
    }
  }
 return 0;
}/* ende von main */
